; Sample routines
; Jan Wassenberg 2000

global init
global check_com
global set_sample_type
global sample


section CONST use16

timer_period		dq	5.49254164988E-2	; BIOS timer frequency

uart			dw	0x03FE			; UART MSR I/O address

convert			dw	cap
convert_table		dw	cap, tem, hum		; Conversion procedures

cap_sub			dq	2.3553E-5
cap_mul			dq	4.3450E+5

tem_sub			dq	0.0000E+0
tem_mul			dq	1.0000E+0

hum_mul			dq	1.0863E+6
hum_sub			dq	3.7758E+2


section DATA use16

delay			resd	1	; Overhead of cpuid & rdtsc
temp			resq	1

cpu_period		resq	1


section CODE use16

;
; get_clocks
; in:
; out:
;   eax - CPU clocks in 55 ms
; destroys:
;   eax, ebx, ecx, edx, esi, es, di
;
get_clocks:
	mov	ax, 0x0040		; BIOS data segment
	mov	es, ax			;
	mov	di, 0x006C		; [es:di] = BIOS tick count
	mov	al, [es:di]		; al = tick count
.1:	cmp	al, [es:di]		; Tick count changed?
	je	.1			; No - wait

	cpuid				; Serialize intruction flow
	rdtsc				; ReaD TimeStamp Counter
;---------------------------------------;
	mov	esi, edx		;
	mov	esi, eax		; Save start clock count
	mov	al, [es:di]		; al = tick count
.2:	cmp	al, [es:di]		; Tick count changed?
	je	.2			; No - wait
					;
	cpuid				; Serialize intruction flow
	rdtsc				; ReaD TimeStamp Counter
;---------------------------------------;

	ret

;
; init
; in:
; out:
; destroys:
;   eax, ebx, ecx, edx, esi
;
init:
	or	al, -1			; Mask all interrupts
	out	0x21, al

%rep 6					; "Warm up" cpuid & rdtsc
	cpuid				; Serialize intruction flow
	rdtsc				; ReaD TimeStamp Counter
;---------------------------------------;
	mov	esi, edx		;
	mov	esi, eax		; Save start clock count
	cpuid				; Serialize intruction flow
	rdtsc				; ReaD TimeStamp Counter
;---------------------------------------;
%endrep

	sub	eax, esi		; Substract start clock count
	mov	[delay], eax		; [Delay] = cpuid, rdtsc & mov overhead

	mov	al, 0xfe		; Mask all interrupts except PIT
	out	0x21, al

%rep 3
	call	get_clocks		; eax = elapsed time
%endrep

	sub	eax, esi		; Subtract start clock count
	sub	eax, [delay]		; Subtract overhead
	mov	[temp], eax		; [Temp] = CPU clocks in 55 ms

	xor	al, al			; Unmask all interrupts
	out	0x21, al

	fld	qword [timer_period]
	fild	dword [temp]
	fdivp	st1, st0		; Clock period = 55 ms / clocks / 55 ms
	fst	qword [cpu_period]	; Save value

	ret

;
; check_com
; in:
; out:
;   al - Cable is attached (Boolean)
; destroys:
;   al, dx
;
check_com:
	; Set DTR line on both UARTs
	mov	al, 0x01		; MCR value (DTR enabled)
	mov	dx, 0x03fc		; COM1 MCR
	out	dx, al
	dec	dh			; COM2 MCR
	out	dx, al

	; Check both UARTs for DSR & DCD == 1
	add	dl, 2			; COM2 MSR
	in	al, dx
	and	al, 0xa0		; Mask bits
	cmp	al, 0xa0		; Both bits set?
	je	.1			; Yes

	inc	dh			; COM1 MSR
	in	al, dx
	and	al, 0xa0		; Mask bits
	cmp	al, 0xa0		; Both bits set?

.1:	sete	al
	mov	[uart], dx		; Store current UART I/O address

	ret

;
; set_sample_type(sample_type : Word);
; destroys:
;   ax, bx
;
set_sample_type:
	push	bp
	mov	bp, sp			; Stack frame

	mov	bx, [bp+4]		; bx = sample type (0..2)
	shl	bx, 1
	mov	ax, [convert_table+bx]	; ax = conversion function address
	mov	[convert], ax		; Save ax

	pop	bp
	ret

;
; get_signal_period
; in:
; out:
;   eax - Elapsed time in CPU clocks
; destroys:
;   eax, ebx, ecx, edx
;
get_signal_period:
	mov	dx, [uart]
	mov	[uart], dx		; Ensure [UART] is cached
.1:	in	al, dx			; UART MSR register
	test	al, 0x04		; Trailing Edge of RI?
	jz	.1			; No - wait

	cpuid				; Serialize intruction flow
	rdtsc				; ReaD TimeStamp Counter
;---------------------------------------;
	mov	edi, edx		;
	mov	esi, eax		; Start clock count
					;
	mov	dx, [uart]		;
.2:	in	al, dx			; UART MSR register
	test	al, 0x04		; Trailing Edge of RI?
	jz	.2			; No - wait

	cpuid				; Serialize intruction flow
	rdtsc				; ReaD TimeStamp Counter
;---------------------------------------;

	ret

;
; sample
; in:
; out:
;   st0 - Value in [convert] units
; destroys:
;
sample:
	cli				; Disable interrupts
%rep 4					; Eliminate cache effects
	call	get_signal_period	; eax = signal period + [Delay]
%endrep
	sti				; Enable interrupts

	sub	eax, esi		; Subtract start clock count
	sub	eax, [delay]		; Subtract overhead
	sub	edx, edi
	mov	[temp  ], eax
	mov	[temp+4], edx

	fild	qword [temp]
	fmul	qword [cpu_period]	; seconds = clocks * clock period

	jmp	[convert]		; Convert RETs back to caller

;
; Conversion routines
; in:
;   st0 - Value to convert
; out:
;   st0 - Converted value
; destroy:
;
cap:
	fsub	qword [cap_sub]		; Value -= 2.3553E-5
	fmul	qword [cap_mul]		; Value *= 4.3450E+5

	ret

tem:
	fsub	qword [tem_sub]		; Value -=
	fmul	qword [tem_mul]		; Value *=

	ret

hum:
	fmul	qword [hum_mul]		; Value *= 1.0863E+6
	fsub	qword [hum_sub]		; Value -= 3.7758E+2

	ret